@@ -2,13 +2,14 @@
 
 import multiprocessing
 import os
+import logging
 from collections import defaultdict
 from typing import Dict
 import numpy as np
 import pandas as pd
 import xarray as xr
 
-
+from cythoninstallhelpers.configure_logger import get_logger
 from cythonarrays.array_properties import _ArrayProperties
 from matrixconverters.save_ptv import SavePTV
 from matrixconverters.xarray2netcdf import xr2netcdf
@@ -20,8 +21,12 @@ from wiver.wiver_cython import (_WIVER,
 np.seterr(divide='ignore', invalid='ignore')
 
 
+class InvalidWiverInputData(ValueError):
+    """Invalid input data for Wiver model"""
+
+
 class WIVER(_WIVER, _ArrayProperties):
-    """WIVER Commercial Trips Model"""
+    """WIVER Business Trip Model"""
     _coordinates = {'n_groups': 'groups',
                     'n_modes': 'modes',
                     'n_zones': 'zone_name',
@@ -32,18 +37,38 @@ class WIVER(_WIVER, _ArrayProperties):
     def __init__(self,
                  n_groups: int,
                  n_zones: int,
-                 n_savings_categories: int=9,
                  n_time_slices: int=5,
                  n_modes: int=4,
                  n_sectors: int=2,
-                 n_threads: int=None):
+                 n_threads: int=None,
+                 logger: logging.Logger=None):
+        """
+        Parameters
+        ----------
+        n_groups:
+            the number of groups
+        n_zones:
+            the number of zones
+        n_time_slices:
+            the number of time slices
+        n_modes:
+            the number of transport modes
+        n_sectors:
+            the number of economic sectors
+        n_threads:
+            the maximum number of threads to use for calculation
+        logger:
+            an optional logger to use from within wiver. If not provided,
+            a logger is created
+        """
         super().__init__()
 
+        self.logger = logger or get_logger(self)
+
         self.n_modes = n_modes
         self.n_groups = n_groups
         self.n_zones = n_zones
         self.n_sectors = n_sectors
-        self.n_savings_categories = n_savings_categories
         self.n_time_slices = n_time_slices
         self.set_n_threads(n_threads=n_threads)
 
@@ -56,7 +81,8 @@ class WIVER(_WIVER, _ArrayProperties):
 
         Parameters
         ----------
-        n_threads: int, optional
+        n_threads:
+            the maximum number of threads to use for calculation
         """
         if n_threads:
             n_threads = min(n_threads, multiprocessing.cpu_count())
@@ -67,9 +93,18 @@ class WIVER(_WIVER, _ArrayProperties):
     @classmethod
     def read_from_netcdf(cls,
                          files: Dict[str, str],
-                         n_threads: int=None) -> 'Wiver':
-        """Read a Wiver Model
-        from a set of netcdf-Filename located in folder"""
+                         n_threads: int=None) -> 'WIVER':
+        """
+        Read a Wiver Model
+        from a set of netcdf-Filename defined in files
+
+        Parameters
+        ----------
+        files:
+            dict with the file names of the input data
+        n_threads:
+            the maximum number of threads to use for calculation
+        """
         # create instance of self
         self = cls(n_groups=0, n_zones=0, n_threads=n_threads)
         # add datasets
@@ -82,7 +117,6 @@ class WIVER(_WIVER, _ArrayProperties):
         self.n_zones = dims['origins']
         self.n_groups = dims['groups']
         self.n_modes = dims['modes']
-        self.n_savings_categories = dims['savings']
         self.n_time_slices = dims['time_slices']
         self.set_n_threads()
 
@@ -108,8 +142,7 @@ class WIVER(_WIVER, _ArrayProperties):
         self.mode_name = ds.mode_name.data
         self.sector_short = ds.sector_short.data
         self.param_dist_g = ds.param_dist.data
-        self.savings_bins_s = ds.savings_bins.data
-        self.savings_weights_gs = ds.savings_weights.data
+        self.savings_param_g = ds.savings_param.data
         self.tour_rates_g = ds.tour_rates.data
         self.stops_per_tour_g = ds.stops_per_tour.data
         self.time_series_starting_trips_gs = ds.time_series_starting_trips.data
@@ -132,7 +165,14 @@ class WIVER(_WIVER, _ArrayProperties):
         self.results = self.define_results()
 
     def set_arrays_from_balancing_ds(self, ds: xr.Dataset):
-        """Set arrays from balancing dataset"""
+        """
+        Set arrays from balancing dataset
+
+        Parameters
+        ----------
+        ds:
+            Dataset with the balancing factors
+        """
         # balancing factors
         self.balancing_factor_gj = ds.balancing_factor.data
         self.trips_to_destination_gj = ds.trips_to_destination.data
@@ -152,10 +192,9 @@ class WIVER(_WIVER, _ArrayProperties):
         self.init_array('mode_g', 'n_groups', 0)
         self.init_array('sector_g', 'n_groups', 0)
         self.init_array('active_g', 'n_groups', 1)
+        self.init_array('converged_g', 'n_groups', 0)
 
-        self.init_array('savings_bins_s', 'n_savings_categories')
-        self.init_array('savings_weights_gs',
-                        'n_groups, n_savings_categories', 1)
+        self.init_array('savings_param_g', 'n_groups', 1.1)
 
         self.init_array('km_ij', 'n_zones, n_zones')
         self.init_array('travel_time_mij', 'n_modes, n_zones, n_zones')
@@ -175,7 +214,7 @@ class WIVER(_WIVER, _ArrayProperties):
                         'n_groups, n_zones, n_zones', 0)
         self.init_array('linking_trips_gij', 'n_groups, n_zones, n_zones', 0)
         self.init_array('return_trips_gij', 'n_groups, n_zones, n_zones', 0)
-        self.init_array('p_destination_tij', 'n_threads, n_zones, n_zones', 0)
+        self.init_array('p_destination_tj', 'n_threads, n_zones', 0)
         self.init_array('p_links_tij', 'n_threads, n_zones, n_zones', 0)
 
         self.init_array('trips_gsij',
@@ -193,6 +232,11 @@ class WIVER(_WIVER, _ArrayProperties):
         self.init_array('time_series_ending_trips_gs',
                         'n_groups, n_time_slices', 1)
 
+    def validate_input_data(self):
+        """make checks on the input data"""
+        if (self.stops_per_tour_g < 1).any():
+            raise InvalidWiverInputData('Stops per Tour < 1.0 are not allowed')
+
     def define_datasets(self):
         """Define the datasets"""
         self.params = self.define_params()
@@ -202,7 +246,9 @@ class WIVER(_WIVER, _ArrayProperties):
         self.results = self.define_results()
 
     def define_params(self) -> xr.Dataset:
-        """Define the params"""
+        """
+        Define the params and return them as dataset
+        """
         ds = xr.Dataset()
         ds['modes'] = self.modes
         ds['mode_name'] = (('modes'), self.mode_name)
@@ -225,10 +271,8 @@ class WIVER(_WIVER, _ArrayProperties):
 
         ds['param_dist'] = (('groups'),
                             self.param_dist_g)
-        ds['savings_bins'] = (('savings'),
-                              self.savings_bins_s)
-        ds['savings_weights'] = (('groups', 'savings'),
-                                 self.savings_weights_gs)
+        ds['savings_param'] = (('groups'),
+                              self.savings_param_g)
         ds['tour_rates'] = (('groups'),
                             self.tour_rates_g)
         ds['stops_per_tour'] = (('groups'),
@@ -242,7 +286,7 @@ class WIVER(_WIVER, _ArrayProperties):
         return ds
 
     def define_balancing(self) -> xr.Dataset:
-        """Define the balancing factors"""
+        """Define the balancing factors and return them as dataset"""
         ds = xr.Dataset()
         ds['balancing_factor'] = (('groups', 'destinations'),
                                     self.balancing_factor_gj)
@@ -251,7 +295,7 @@ class WIVER(_WIVER, _ArrayProperties):
         return ds
 
     def define_zonal_data(self) -> xr.Dataset:
-        """Define the matrices"""
+        """Define the zonal data and return them as dataset"""
         ds = xr.Dataset()
         ds['groups'] = self.groups
         ds['zone_no'] = self.zone_no
@@ -265,7 +309,7 @@ class WIVER(_WIVER, _ArrayProperties):
         return ds
 
     def define_matrices(self) -> xr.Dataset:
-        """Define the matrices"""
+        """Define the matrices and return them as dataset"""
         ds = xr.Dataset()
         ds['modes'] = self.modes
         ds['origins'] = self.zone_no
@@ -280,7 +324,7 @@ class WIVER(_WIVER, _ArrayProperties):
         return ds
 
     def define_results(self) -> xr.Dataset:
-        """Define the results"""
+        """Define the results and return them as dataset"""
         ds = xr.Dataset()
         # resulting arrays
         ds['modes'] = self.modes
@@ -313,7 +357,14 @@ class WIVER(_WIVER, _ArrayProperties):
                               self.results))
 
     def save_all_data(self, wiver_files: Dict[str, str]):
-        """Save Dataset to netcdf-file"""
+        """
+        Save Dataset to netcdf-files
+
+        Parameters
+        ----------
+        wiver_files:
+            dict with the input and output filenames
+        """
         datasets = ('params', 'zonal_data', 'matrices',
                     'balancing', 'results')
         for dataset_name in datasets:
@@ -321,7 +372,16 @@ class WIVER(_WIVER, _ArrayProperties):
             self.save_data(dataset_name, fn)
 
     def save_data(self, dataset_name: str, fn: str):
-        """Save Dataset to netcdf-file"""
+        """
+        Save Dataset to netcdf-file
+
+        Parameters
+        ----------
+        dataset_name:
+            the dataset that should be saved
+        fn:
+            filename to write to
+        """
         ds = getattr(self, dataset_name)
         self.logger.info('write {}'.format(fn))
         dirname = os.path.dirname(fn)
@@ -331,21 +391,44 @@ class WIVER(_WIVER, _ArrayProperties):
         ds.close()
 
     def read_all_data(self, datasets: Dict[str, str]):
-        """Read Datasets from netcdf-file"""
+        """
+        Read Datasets from netcdf-file
+
+        Parameters
+        ----------
+        datasets:
+            dict with the filenames of the input and output data
+        """
         for dataset_name in ('params', 'matrices', 'zonal_data' ,
                              'balancing'):
             fn = datasets[dataset_name]
             self.read_data(dataset_name, fn)
 
     def read_data(self, dataset_name: str, fn: str):
-        """read single dataset from """
+        """
+        read single dataset from file
+
+        Parameters
+        ----------
+        dataset_name:
+            the dataset that should be read
+        fn:
+            filename to read from
+        """
         self.logger.info('read {}'.format(fn))
         ds = xr.open_dataset(fn).load()
         ds.close()
         setattr(self, dataset_name, ds)
 
     def save_results(self, wiver_files: Dict[str, str]):
-        """Save results except trips_gsij to folder"""
+        """
+        Save results except trips_gsij to folder
+
+        Parameters
+        ----------
+        wiver_files:
+            dict with the filenames of the input and output data
+        """
         del self.results['trips_gsij']
         dataset_name = 'results'
         fn = wiver_files[dataset_name]
@@ -355,7 +438,16 @@ class WIVER(_WIVER, _ArrayProperties):
         self.save_data(dataset_name, fn)
 
     def save_results_to_visum(self, folder: str, visum_format: str='BK'):
-        """Save the results to VISUM-Format"""
+        """
+        Save the results to VISUM-Format
+
+        Parameters
+        ----------
+        folder:
+            the folder where the Visum-Matrices will be written
+        visum_format:
+            the Visum-Matrix format (Default=Binary Compressed (BK))
+        """
         for m, mode in enumerate(self.modes):
             visum_ds = xr.Dataset()
             visum_ds['matrix'] = self.results.trips_mij[m]
@@ -373,7 +465,16 @@ class WIVER(_WIVER, _ArrayProperties):
     def save_detailed_results_to_visum(self,
                                        folder: str,
                                        visum_format: str='BK'):
-        """Save the results to VISUM-Format"""
+        """
+        Save the detailed results to VISUM-Format including the matrices for each group
+
+        Parameters
+        ----------
+        folder:
+            the folder where the Visum-Matrices will be written
+        visum_format:
+            the Visum-Matrix format (Default=Binary Compressed (BK))
+        """
         sectors = defaultdict(list)
         for g, group in enumerate(self.groups):
             if self.active_g[g]:
@@ -416,43 +517,83 @@ class WIVER(_WIVER, _ArrayProperties):
             s.savePTVMatrix(file_name, file_type=visum_format)
             self.logger.info('matrix_saved')
 
-    def adjust_balancing_factor(self, threshold: float=0.1):
+    def adjust_balancing_factor(self, iteration: int, threshold: float=0.1):
         """
-        Randsummenabgleich für Zielpotenziale
+        adjust the balancing factor for the destination choice model
+
+        Parameters
+        ----------
+        threshold:
+            if the difference between the modelled and the target trips is < threshold,
+            the model converges
         """
-        self.converged=False
+        not_converged = ~ self.converged_g.astype(bool)
         sp = self.zonal_data.sink_potential
         target_share = sp / sp.sum('destinations')
         trips = self.balancing.trips_to_destination
         actual_share = trips / trips.sum('destinations')
         kf = target_share / actual_share
-        kf[:] = kf.fillna(1)
+        kf[not_converged] = kf[not_converged].fillna(1)
         bf = self.balancing.balancing_factor
         # adjust balancing factor
-        bf[:] = bf.fillna(1)
-        bf[:] = bf * kf
+        bf[not_converged] = bf[not_converged].fillna(1)
+        bf[not_converged] = bf[not_converged] * kf[not_converged]
         # normalize balancing factor
-        #bf[:] = bf / bf.mean('destinations')
-        if (np.abs(kf - 1) < threshold).all():
-            self.converged = True
-            self.logger.info('converged!')
+        bf[not_converged] = bf[not_converged] / bf.mean('destinations')[not_converged]
+        for g, group in enumerate(self.groups):
+            if not_converged[g]:
+                if (np.abs(kf[g] - 1) < threshold).all():
+                    self.converged_g[g]= iteration
+                    self.logger.info(
+                        f'group {group} converged after {iteration} iterations')
+                else:
+                    self.logger.debug(f'group {group} not converged yet')
+            else:
+                self.logger.debug(f'group {group} was already converged')
+
+    def calc(self):
+        """calc the daily trips and the trips by time slice"""
+        for g in range(self.n_groups):
+            if self.converged_g[g] or not self.active_g[g]:
+                continue
+            self.calc_daily_trips(g)
+            self.adjust_linking_trips(g)
+            self.trips_gij[g] = (self.home_based_trips_gij[g] +
+                                self.linking_trips_gij[g] +
+                                self.return_trips_gij[g])
+            self.trips_to_destination_gj[g] = (self.home_based_trips_gij[g] +
+                                               self.linking_trips_gij[g]).sum(0)
+        self.calc_time_series()
+        self.aggregate_to_modes()
+        self.calc_mean_distance()
+        self.calc_mean_distance_mode()
 
     def calc_with_balancing(self,
                             max_iterations: int=10,
                             threshold: float=0.1):
-        """calculate with balancing the """
-        self.converged = False
+        """
+        calculate with balancing the destination choice model
+
+        Parameters
+        ----------
+        max_iterations:
+            the maximum number of iterations
+        threshold:
+            if the difference between the modelled and the target trips is < threshold,
+            the model converges
+        """
+        self.converged_g[:] = 0
         iteration = 0
-        while (not self.converged) and iteration < max_iterations:
+        while (not self.converged_g.all()) and iteration < max_iterations:
             iteration += 1
-            self.logger.info('calculate trips in iteration {}'.format(iteration))
+            self.logger.info(f'calculate trips in iteration {iteration}')
             self.calc()
-            self.logger.info('Total trips: {:0.2f}'.format(self.trips_gij.sum()))
-            self.adjust_balancing_factor(threshold)
+            self.logger.info(f'Total trips: {self.trips_gij.sum():0.2f}')
+            self.adjust_balancing_factor(iteration, threshold)
 
     def calc_starting_and_ending_trips(self) -> pd.DataFrame:
         """
-        calculate the starting trips per zone and group
+        calculate the starting trips per zone and group and return them as Dataframe
         """
         trips_ij = self.results.trips_gij.sum('groups')
         modelled_trips = (trips_ij.sum('origins').\
